AUTOJINJA_DEFAULT_FILENAME   = "__jinja__.py"
AUTOJINJA_DEFAULT_TAG        = "autojinja"
AUTOJINJA_DEFAULT_COG_OPEN   = "[[["
AUTOJINJA_DEFAULT_COG_CLOSE  = "]]]"
AUTOJINJA_DEFAULT_COG_END    = "end"
AUTOJINJA_DEFAULT_EDIT_OPEN  = "<<["
AUTOJINJA_DEFAULT_EDIT_CLOSE = "]>>"
AUTOJINJA_DEFAULT_EDIT_END   = "end"

AUTOJINJA_REMOVE_MARKERS = "AUTOJINJA_REMOVE_MARKERS"
AUTOJINJA_SILENT         = "AUTOJINJA_SILENT"
AUTOJINJA_THIS_DIRPATH   = "THIS_DIRPATH"
